/*
 * Copyright (C) 1999 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:      Dean Johnson (dtj@sgi.com)
 *
 * Purpose:     Handles the creation and maintainence of the connection
 *		with the nexxus.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib.h>
#include "node.h"
#include <arpa/inet.h>
#include "vacmclient_api.h"
#include "nexxus_connection.h"
#include "input_queue.h"
#include "node_list.h"
#include "hoover.h"
#include "dialogs.h"

extern void	setInputTag(void *);
extern char	*getNexxusData(void *);
extern void	textScan(char *);

static void
dialog_destroyed(GtkWidget *w,gpointer data)
{
	ConnectInfo *ci = (ConnectInfo *) data;

	/*
	 *	The informational dialog is being destroyed through
	 *	typical gtk means, so null out our pointer to it so 
	 *	that it doesn't try to update it.
	 */
	ci->dialog = NULL;
}

int
ConnectInfo::makeConnection(char *theHost,char *id,char *pw,
			    gboolean suppress_err)
{
	char	*inPtr;
	void	*theHandle;
	int	ret;
	char	buffer[512];

	g_return_val_if_fail(theHost!=NULL,0);
	g_return_val_if_fail(*theHost!='\0',0);

	/*--------------------------------------------------------------*
	 *		Connect and wait for it to be "cooked"		*
	 *--------------------------------------------------------------*/

	if (id == NULL && pw == NULL) {
		// just use the defaults
		theUserid   = strdup((id==NULL) ? "blum" : id);
		thePassword = strdup((pw==NULL) ? "frub" : pw);
		ret = api_nexxus_connect(theHost,"blum","frub",&theHandle);
	}
	else	{
		if (id) theUserid = strdup(id);
		if (pw) thePassword = strdup(pw);
		ret = api_nexxus_connect(theHost,id,pw,&theHandle);
	}
        if (ret != API_RETURN_OK && ret != API_RETURN_ALREADY_CONNECTED) {
		if (!suppress_err) {
			sprintf(buffer,"Cannot connect to `%s`!\nAttempt: %d",
				theHost,attempt);
			if (attempt == 1) {
				dialog = informational_dialog(buffer);
				gtk_signal_connect(GTK_OBJECT(dialog),"destroy",
                                  	GTK_SIGNAL_FUNC(dialog_destroyed),
	                                (GtkObject *) this);

			}
			else if (dialog != NULL) {
				informational_dialog_update(dialog,buffer);
			}
//			api_nexxus_perror("makeConnection()");
		}
		return(0);
	}
	handle = theHandle;
	while (1) {
		if (!(inPtr = getNexxusData(handle))) {
			if (!suppress_err) {
				fprintf(stderr,"ConnectInfo::makeConnection() - getNexxusData(handle) failed!\n");
			}
			return(0);
		}
                if (strstr(inPtr,"NEXXUS_READY")) {
			free(inPtr);	/* clean up after ourselves */
			break;
		}
	}
	setInputTag(handle);
	connected = TRUE;

	/*--------------------------------------------------------------*
	 *		Send the SET_DEMO_FLAG request			*
	 *--------------------------------------------------------------*/

	if (globalDemoMode) {
		(void) sendMsg("emp:SET_DEMO_FLAG:on:",TRUE,NULL,(gpointer) NULL);
	}
	/*
	 *	make sure that it tells me lots of interesting stuff
	 */
	(void) sendMsg("nexxus:IPC_CHATTY:on:",TRUE,NULL,(gpointer) NULL);

	refreshNodeList();	// get the list of nodes for this nexxus
	return(1);
}

Node *
handleToRootNode(void *handle)
{
	Node *tmp;

	g_return_val_if_fail(handle!=NULL,NULL);
	for (unsigned int i = 0; i < g_slist_length(theNodeList); i++) {
		tmp = (Node *) g_list_nth_data((GList *) theNodeList,i);
		if (tmp->singularTypeP()) continue;	// not a leaf node
		if (tmp->connection->handle != handle) continue;
		return(tmp);
	}
	return(NULL);
}

ConnectInfo *
handleToConnectInfo(void *handle)
{
	Node	*tmp;

	g_return_val_if_fail(handle!=NULL,NULL);
	for (unsigned int i = 0; i < g_slist_length(theNodeList); i++) {
		tmp = (Node *) g_list_nth_data((GList *) theNodeList,i);
		if (tmp->connection->handle == handle)
			return(tmp->connection);
	}
	return(NULL);
}

int
numClusters()
{
	int	cnt = 0;
	Node	*ctmp;

	for (unsigned int i = 0; i < g_slist_length(theNodeList); i++) {
		ctmp = (Node *) g_list_nth_data((GList *) theNodeList,i);
		if (ctmp->type == TYPE_CLUSTER) 
			cnt++;
	}
	return(cnt);
}

void
ConnectInfo::refreshNodeList()
{
	if (!connected) {
		// try to reconnect
		connect();
	}
	if (handle != NULL) {
		/* just do the specified cluster, rather than all of them */
		sendMsg("NEXXUS:NODE_VAR_LIST:TYPE:",TRUE,handleNodeListMessage,
			(gpointer) handle);
		return;
	}
}

/*----------------------------------------------------------------------*
 *			     IPC MESSAGE SENDER				*
 *----------------------------------------------------------------------*/

unsigned int
sendIPCMsg(void *theHandle,char *cmd,gboolean atCompletion,
	   void (*theFunc)(InputEntry *data,gpointer clientData),
           gpointer cd)
{
	ConnectInfo	*connect;

	g_return_val_if_fail(theHandle!=NULL,0);
	if (globalVerboseMode) {
		fprintf(stderr,"sendIPCMsg(\"%s\",%d,0x%08x)\n",cmd,
			atCompletion,(unsigned int) theFunc);
	}
	connect = handleToConnectInfo(theHandle);
	if (connect != NULL && !connect->connected) {
		// don't even send it because it isn't connect
		fprintf(stderr,"Can't send message to '%s'! Trying to reconnect.\n",connect->hostname);
		connect->connect();
		return(0);
	}
        api_nexxus_send_ipc(theHandle,cmd,strlen(cmd));
	queueJobEntry(cmd,atCompletion,theFunc,cd);
	return(0);
}

Node *
ConnectInfo::renderClusterNode(char *userid,char *password)
{
        char    buffer[512];

        /*
         *      Put together the node.
         */
        sprintf(buffer,"Cluster %s",hostname);
        node = new Node(TYPE_CLUSTER,buffer,NULL,this);
	gtk_tree_append(GTK_TREE(theList), node->node_item);
	gtk_widget_show(node->node_item);
        /*
         *      Actually try and make the connection now
         */
	if (!connected)
		connect(5,userid,password);
	return(node);
}

