/*
 * Copyright (C) 1999 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:	Dean Johnson (dtj@sgi.com)
 *
 * Purpose:	These are the various actions that can be done within hoover.
 *
 */

#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include "node.h"
#include "hoover.h"
#include "HooverControl.h"
#include "SENSListPanel.h"
#include "SELListPanel.h"
#include "MainListPanel.h"
#include "PropListPanel.h"
#include "ConsolePanel.h"
#include "dialogs.h"
#include "libgessie/Panel.h"
#include "libgessie/Globals.h"
#include "libgessie/Gessie.h"
#include "libgessie/Broker.h"

extern void	createProperties(Node *,gpointer data);
extern MainListPanel *mainList;
extern int	checkTreeStates(gpointer data);	// hoover.C

void
refreshNodeLists()
{
	Node    *tmp;

	/* do all the clusters */
	for (unsigned int i = 0; i < g_slist_length(theNodeList); i++) {
		tmp = (Node *) g_list_nth_data((GList *) theNodeList,i);
		tmp->connection->refreshNodeList();
	}
}

void
suspendStatusChecks()
{
	globalStatusChecksSuspended = TRUE;
}

void
resumeStatusChecks()
{
	globalStatusChecksSuspended = FALSE;
	checkTreeStates(NULL);  // make sure everything is updated
}

static void
doReset(Node *n,gpointer /*data*/)
{
	n->services->doReset();
}

static void
doPowerDown(Node *n,gpointer /*data*/)
{
	n->services->doPowerDown();
}

static void
doPowerUp(Node *n,gpointer /*data*/)
{
	n->services->doPowerUp();
}

static  void
createSensorLog(Node *n,gpointer /*data*/)
{
	char	buf[1024];
	Panel *p = new SENSListPanel(theRootNotebook,n);
	HooverControl *hoov = (HooverControl *) gsSearchPanelsUpward(mainList,
							"HooverControl");
	sprintf(buf,"%s:Sensor Log",n->name);
	if (hoov == NULL) {
		fprintf(stderr,"cannot find HooverControl for SEL placement\n");
		theRootNotebook->add(p,buf);
	}
	else	{
		hoov->addBot(p,buf);
	}
}

static  void
createLog(Node *n,gpointer /*data*/)
{
	/*
	 *	check to see if it exists already
	 */
	GSList *plist = gsGetPanelList(mainList,"SELListPanel");
	if (plist != NULL) {
		for (unsigned int i = 0; i < g_slist_length(plist); i++) {
			SELListPanel *sp = (SELListPanel *) g_list_nth_data((GList *) plist,i);
			if (sp->theNode == n) {
				//	raise it and return;
				sp->raise();
				return;
			}
		}
	}
	char	buf[1024];
	EventLogType	evl = (n->type == TYPE_RSH_NODE) ? EV_LOG_TEXT : EV_LOG_SEL;
	Panel *p = new SELListPanel(theRootNotebook,n,evl);
	HooverControl *hoov = (HooverControl *) gsSearchPanelsUpward(mainList,
							"HooverControl");
	sprintf(buf,"%s:SEL Log",n->name);
	if (hoov == NULL) {
		fprintf(stderr,"cannot find HooverControl for SEL placement\n");
		theRootNotebook->add(p,buf);
	}
	else	{
		hoov->addBot(p,buf);
	}
}

static void 
createConsole(Node *n,gpointer data)
{ 
#define	LAST_RESORT
#ifdef	LAST_RESORT
	/*
	 *	This code is a pathetic attempt to get some sort
	 *	of working console on the EMP and MSC nodes. It is
	 *	stolen nearly intact from flim.
	 */
	if (n->type != TYPE_RSH_NODE) {
		if (!fork()) {
			char buf[4096];

			sprintf(buf,"xterm -T \"%s:Console \" -e sercon_terminal -n \"%s\" -u \"%s\" -p \"%s\" -i \"%s\" -m RW",
				n->name,
				n->connection->hostname,
				n->connection->theUserid,
				n->connection->thePassword,
				n->name);
			execl("/bin/sh", "/bin/sh", "-c", buf, NULL);
			exit(0);
		}
		return;
	}
#endif	/*LAST_RESORT*/

	/*
	 *	check to see if it exists already
	 */
	GSList *plist = gsGetPanelList(mainList,"ConsolePanel");
	if (plist != NULL) {
		for (unsigned int i = 0; i < g_slist_length(plist); i++) {
			ConsolePanel *sp = (ConsolePanel *) g_list_nth_data((GList *) plist,i);
			if (sp->theNode == n) {
				//	raise it and return;
				sp->raise();
				return;
			}
		}
	}
	char	buf[1024];
	Panel *p = new ConsolePanel(n,data);
	HooverControl *hoov = (HooverControl *) gsSearchPanelsUpward(mainList,
							"HooverControl");
	sprintf(buf,"%s:Console",n->name);
	if (hoov == NULL) {
		fprintf(stderr,"cannot find HooverControl for Console placement\n");
		theRootNotebook->add(p,buf);
	}
	else	{
		hoov->addBot(p,buf);
	}
}

#define	NODE_FUNC(nm,func)				\
	static void nm(Node *n,gpointer data)		\
	{ 						\
		n->mapFunc(func,data);			\
	}

NODE_FUNC(showLog,createLog);
NODE_FUNC(showConsole,createConsole);
NODE_FUNC(showProperties,createProperties);
NODE_FUNC(showSensors,createSensorLog);

static void 
removeNode(Node *n,gpointer /*data*/)
{ 
	delete n;
        refreshNodeLists();
}

class	ActionSquirrel {
public:
	ActionSquirrel(int act,Node *n) {
		action = act;
		node   = n;
	};
	int	action;
	Node	*node;
};

void
reallyDoAction(gpointer data,GtkWidget *) 
{
	ActionSquirrel *tmp = (ActionSquirrel *) data;
	int	action = tmp->action;
	Node	*node = tmp->node;

	delete  tmp;	// clean me up before going on.

	switch (action) {
	case RESET_IT:		node->mapFunc(doReset);		break;
	case POWER_DOWN:	node->mapFunc(doPowerDown);	break;
	case POWER_UP:		node->mapFunc(doPowerUp);	break;
	case VIEW_LOG:		node->mapFunc(showLog);		break;
	case SERIAL_CONSOLE:	node->mapFunc(showConsole);	break;
	case REMOVE_IT:		node->mapFunc(removeNode);	break;
	case SENSE_IT:		node->mapFunc(showSensors);	break;
	case PROPERTIES:	
		if (node->singularTypeP()) {
			node->mapFunc(showProperties);
		}
		else	{
			/*
			 *	check to see if it exists already
			 */
			GSList *plist = gsGetPanelList(mainList,"PropListPanel");
			if (plist != NULL) {
				for (unsigned int i = 0; i < g_slist_length(plist); i++) {
					PropListPanel *sp = (PropListPanel *) g_list_nth_data((GList *) plist,i);
					if (sp->theNode == node) {
						//	raise it and return;
						sp->raise();
						return;
					}
				}
			}

			Panel *p = new PropListPanel(theRootNotebook,node);
			HooverControl *hoov = (HooverControl *) gsSearchPanelsUpward(mainList,
									"HooverControl");
			if (hoov == NULL) {
				fprintf(stderr,"cannot find HooverControl for SEL placement\n");                theRootNotebook->add(p,"Properties List");
			}
			else    {
				hoov->addBot(p,"Properties List");
			}

		}
		break;
	}
}

void
doNodeAction(gpointer data,GtkWidget *w)
{
	int	action;
	Node	*node;

	if (w == NULL) return;
	action = (int) data;
	if (!mainList->currentSelection()) {
		/* nothing to work on */
		informational_dialog("No Node Selected!");
		return;
	}
	node = (Node *) gtk_object_get_data(GTK_OBJECT(mainList->currentSelection()),"item");
	assert(node != NULL);

	ActionSquirrel *tmp = new ActionSquirrel(action,node);

	/* 
	 *	Danger Will Robinson! Danger! Danger! 
	 *
	 *	Query user to make sure they are happy with their selection.
	 */
	if (   (action == RESET_IT)   || (action == REMOVE_IT)
	    || (action == POWER_DOWN) || (action == POWER_UP)
	   ) {
		char	buffer[1024];
		int	cnt;
		cnt = node->numChildren();
		if (cnt > 1) {
			sprintf(buffer,"Do you really, really want to  \ndo this to all %d nodes?",cnt);
		}
		else {
			sprintf(buffer,"Do you really, really want to do this?");
		}
		binary_dialog(buffer,GTK_SIGNAL_FUNC(reallyDoAction),NULL,
			      (gpointer) tmp);
		
		return;
	}
	reallyDoAction((gpointer) tmp,w);
}


