/*
 * Copyright (C) 1999 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:      Dean Johnson (dtj@sgi.com)
 *
 * Purpose:     This handles the job queues and the input scanning.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <arpa/inet.h>
#include "vacmclient_api.h"
#include "nexxus_connection.h"
#include "input_queue.h"
#include "node.h"
#include "hoover.h"

class JobEntry {
public:
	JobEntry(int jNum,gboolean atC,void (*tFunc)(InputEntry *data,gpointer clientData),gpointer cd)	{ 
			jobNum = jNum;
			atComp = atC;
			callback = tFunc;
			clientData = cd;
			root = NULL;
			last = NULL;
			}
	gpointer	clientData;
	char		*cmd;		// the cmd used
	int		jobNum;
	gboolean	atComp;		// call callback an completion
	/* function to call when something needs to be done */
	void	(*callback)(InputEntry *data,gpointer clientData);

	InputEntry	*root;
	InputEntry	*last;		// last entry in set
};

#define	QUEUED_ENTRY	-2

static int	completeJobEntry(unsigned int jobNum);
static void	addInputEntry(gpointer handle,unsigned int jobNum,char *msg);
static JobEntry	*findQueuedJobEntry(char *cmd);
static JobEntry *findJobEntry(int jobNum);

static GSList	*jobEntryRoot = NULL;

/*----------------------------------------------------------------------*
 *			    HANDLE INPUT QUEUES				*
 *----------------------------------------------------------------------*/

static void
textScan(gpointer handle,char *str)
{
	unsigned int jobnum;

	if (globalVerboseMode >= 3) {
		fprintf(stderr,"textScan-<%s>\n",(char *) str);
	}
	/*
	 *	Do some fiddling with the line
	 */
	(void) strtok(str,":");	/* service */
	char	*tjobnum = strtok(NULL,":");
	char	*msg = (tjobnum == NULL) ? ((char *) "<BAD_MSG>")
					 : tjobnum + strlen(tjobnum) + 1;
	if (tjobnum == NULL || *tjobnum == '\0') {
		fprintf(stderr,"bad message <%s>\n",msg);
		return;	// just bail
	}
	jobnum  = (unsigned int) atoi(tjobnum);

	if (!strncmp("JOB_STARTED",msg,11)) {
		JobEntry *entry;
		char *cp = msg + strlen("JOB_STARTED");
		if (*cp == ':' && (entry = findQueuedJobEntry(cp+1))) {
			// just let it set it
		}
		else	{
			/*
			 *	Just use the first queued entry
			 */
			entry = findJobEntry(QUEUED_ENTRY);
		}
		entry->jobNum = jobnum;
	}
	else if (!strcmp("JOB_COMPLETED",msg)) {
		completeJobEntry(jobnum);
	}
#ifndef	notdef
	else if (!strncmp("JOB_ERROR",msg,9)) {
		addInputEntry(handle,jobnum,msg);
		completeJobEntry(jobnum);
	}
#endif	/*notdef*/
	else	{
		//	its a message
		addInputEntry(handle,jobnum,msg);
	}
}

static  void
inputScan(gpointer data,gint /*source*/,GdkInputCondition /*cond*/)
{
	void	*theHandle = (void *) data;
	char	*buffer;
	unsigned int len;

	if (api_nexxus_recv_ipc(theHandle, &buffer, &len) == API_RETURN_OK) {
		if (globalVerboseMode >= 2) 
			fprintf(stderr,"inputScan received <%s>\n",buffer);
		textScan(theHandle,buffer);
		free(buffer);
	}
}

void
setInputTag(void *theHandle)
{
	NEXXUS_CONNECTION *nexxus = (NEXXUS_CONNECTION *) theHandle;

	gdk_input_add((nexxus->I_fd),GDK_INPUT_READ,inputScan,theHandle);
}

/*----------------------------------------------------------------------*
 *			INPUT PARSING AND QUEUEING			*
 *----------------------------------------------------------------------*/

static JobEntry *
findQueuedJobEntry(char *cmd)
{
	JobEntry *job;
	if (jobEntryRoot == NULL) return(NULL);
	for (unsigned int i = 0; i < g_slist_length(jobEntryRoot); i++) {
		job = (JobEntry *) g_slist_nth_data(jobEntryRoot,i);
		if (job == NULL) continue;
		if (job->jobNum != QUEUED_ENTRY) continue;
		if (!strcmp(job->cmd,cmd)) { 
			return(job);
		}
	}
	return(NULL);
}

static JobEntry *
findJobEntry(int jobNum)
{
	JobEntry *job;
	if (jobEntryRoot == NULL) return(NULL);
	for (unsigned int i = 0; i < g_slist_length(jobEntryRoot); i++) {
		job = (JobEntry *) g_slist_nth_data(jobEntryRoot,i);
		if (job == NULL) continue;
		if (job->jobNum == jobNum) return(job);
	}
	return(NULL);
}

void
newJobEntry(unsigned int jobNum,gboolean atCompletion,
	    void (*theFunc)(InputEntry *data,gpointer clientData),
	    gpointer cd)
{
	/*
	 *	Look to see if it is already registered
	 */
	JobEntry *job = findJobEntry(jobNum);
	if (job != NULL) {
		fprintf(stderr,"job number duplication (%d)!\n",jobNum);
		return;
	}
	job = new JobEntry(jobNum,atCompletion,theFunc,cd);
	job->cmd = NULL;
	if (!jobEntryRoot) {
		jobEntryRoot = g_slist_alloc();
	}
	jobEntryRoot = g_slist_append(jobEntryRoot,job);
}

void
queueJobEntry(char *cmd, gboolean atCompletion,
	      void (*theFunc)(InputEntry *data,gpointer clientData),gpointer cd)
{
	JobEntry *job = new JobEntry(QUEUED_ENTRY,atCompletion,theFunc,cd);
	job->cmd  = strdup(cmd);
	if (!jobEntryRoot) {
		jobEntryRoot = g_slist_alloc();
	}
	jobEntryRoot = g_slist_append(jobEntryRoot,(gpointer) job);
}

int
delJobEntry(unsigned int jobNum)
{
	JobEntry *job = NULL;
	InputEntry	*inp;
	InputEntry	*nxt;

	if (!(job = findJobEntry(jobNum))) { return(0); /* couldn't find it */ }

	g_slist_remove(jobEntryRoot,(gpointer) job);

	/*
	 *	start by nuking the input entries, because the job entry
	 *	doesn't have any of its own storage.
	 */
	inp = job->root;
	while (inp) {
		nxt = inp->next;
		if (inp->string) free(inp->string);
		free(inp);
		inp = nxt;
	}
	/*
	 *	Finally, nuke myself.
	 */
	free(job);
	return(1);
}

static int
completeJobEntry(unsigned int jobNum)
{
	JobEntry *job = findJobEntry(jobNum);
	if (job == NULL) {
		fprintf(stderr,"completeJobEntry(%d): couldnt find job!\n",jobNum);
		return(0);
	}
	/*
	 *	Fire off the completion callback
	 */
	if (globalVerboseMode >= 4) {
		fprintf(stderr,"completeJobEntry - calling jobNum=%d callback 0x%08x\n",
			jobNum,(unsigned int) job->callback);
	}
	if (job->callback != NULL) {
		job->callback(job->root,job->clientData);
	}
	/*
	 *	Do posthumous stuff
	 */
	delJobEntry(jobNum);	// nuke myself
	return(1);
}

static void
addInputEntry(gpointer handle,unsigned int jobNum,char *msg)
{
	InputEntry	*tmp;
	JobEntry *job = findJobEntry(jobNum);
	if (job == NULL) {
		fprintf(stderr,"newInputEntry(%d): couldnt find job!\n",jobNum);
		return;
	}

	tmp = (InputEntry *) malloc(sizeof(InputEntry));
	tmp->jobnum = jobNum;
	tmp->string = strdup(msg);
	tmp->handle = handle;
	tmp->next   = NULL;

	if (!job->atComp && job->callback != NULL) {
		/*
		 *	Don't wait, send it now
		 */
		job->callback(tmp,job->clientData);
		/*
		 *	clean things up
		 */
		free(tmp->string);
		free(tmp);
		return;
	}

	if (job->last == NULL) {
		job->root = job->last = tmp;
	}
	else	{
		job->last->next = tmp;
		job->last = tmp;
	}
}

void
printInputEntries(InputEntry *entry,char *preface)
{
	for (InputEntry *tmp = entry; tmp != NULL; tmp = tmp->next) {
		if (preface) fprintf(stderr,"%s",preface);
		fprintf(stderr,"%s\n",tmp->string);
	}
}

int
countInputEntries(InputEntry *entry)
{
	int	i = 0;

	for (InputEntry *tmp = entry; tmp != NULL; tmp = tmp->next) i++;
	return(i);
}

char *
getNexxusData(void *handle)
{
        char    *inPacket;
        unsigned int     inPacketLen;
	char	*thePacket;
	int	ret;
	ConnectInfo *connect;

	if ((ret = api_nexxus_wait_for_data(handle,&inPacket,&inPacketLen,10)) >= 0) {
		/* copy and null terminate */
		thePacket = (char *) malloc(inPacketLen+1);
		if (thePacket == NULL) {
fprintf(stderr,"getNexxusData() - couldn't get malloc(%d)\n",inPacketLen);
			return(NULL);
		}
		memcpy(thePacket,inPacket,inPacketLen);
		thePacket[inPacketLen] = '\0';
		free(inPacket);		/* clean up incoming packet */
		return(thePacket);
	}
	if (ret == API_RETURN_TIMED_OUT) {
		fprintf(stderr,"input_queue: Nexxus connection reply timed out!\n");
	}
	else if (ret == API_RETURN_NEXXUS_DISCONNECT) {
		// we have gotten disconnected somehow
		connect = handleToConnectInfo(handle);
		if (connect != NULL) {
			connect->connected = FALSE;
			api_nexxus_disconnect(handle);
			fprintf(stderr,"host `%s` has been disconnected!\n",
				connect->hostname);
		}
	}
	else    {
		api_nexxus_perror("input_queue");
	}
	return(NULL);
}
/*
 *	JobEntry -	one entry for each job currently going on.
 */

