/*
 * Copyright (C) 1999 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:	Dean Johnson (dtj@sgi.com)
 *
 * Purpose:	The PanelContainer is a key portion of the infrastructure
 *		that libgessie provides. It, as you might suspect, is a
 *		container for Panels. It also handles the drag and drop
 *		of Panels.
 *
 * Todo:	Add splitting capabilities.
 *
 */

#include <stdio.h>
#include <gdk/gdkx.h>
#include "Gessie.h"
#include "Globals.h"
#include "images/PanelDrag.xpm"
#include "images/tab.xpm"

#define	GB_DRAG_NONE	0

static gboolean drop_rcv_callback(GtkWidget *w,GdkDragContext *con,gint x,gint y,guint time);
static gint	check_children_flags(GtkNotebook *notebook,
				     GtkNotebookPage *page,gint page_num,
                     		     gpointer user_data);


/*
 *	If there are multiple children, make sure there are tabs
 *	visible.
 */

static void
checkOnTabs(GtkNotebook * /*notebook*/,GtkNotebookPage * /*page*/,
		     gint /*page_num*/,gpointer data)
{
        GtkWidget *nb = (GtkWidget *) data;
        GList *children = gtk_container_children((GtkContainer *) nb);
        if (g_list_length(children) > 1) {
                gtk_notebook_set_show_tabs(GTK_NOTEBOOK(nb),TRUE);
        }
}

PanelContainer::PanelContainer() 
{
	parent = NULL;
	theBaseWidget = gtk_notebook_new( );
	gtk_object_set_data((GtkObject *) theBaseWidget,"panelcontainer",
			    (gpointer) this);
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(theBaseWidget),GTK_POS_TOP); 
	gtk_notebook_set_tab_border(GTK_NOTEBOOK(theBaseWidget),0);
	gtk_widget_show(theBaseWidget);
	gtk_drag_dest_set(theBaseWidget,GTK_DEST_DEFAULT_ALL,
			  panel_container_target_table, 
			  n_panel_container_targets,GDK_ACTION_COPY);
	gtk_signal_connect(GTK_OBJECT(theBaseWidget),"drag_drop",
                      	   GTK_SIGNAL_FUNC(drop_rcv_callback),
			   (gpointer) this);
	gtk_signal_connect(GTK_OBJECT(theBaseWidget),"switch-page",
                      	   GTK_SIGNAL_FUNC(check_children_flags),
			   (gpointer) this);
        gtk_signal_connect(GTK_OBJECT(theBaseWidget), "switch-page" /*"add"*/,
                           GTK_SIGNAL_FUNC(checkOnTabs),
			   (gpointer) theBaseWidget);

	setTabBorder(0);
}

static gboolean
idleNukePanel(gpointer data)
{
	Panel *p = (Panel *) data;

	if (p != NULL)
		p->remove();
	return(FALSE);
}

static gint
check_children_flags(GtkNotebook *notebook,GtkNotebookPage * /*page*/,
		     gint page_num,gpointer /*user_data*/)
{
	Panel	*p;

	GList *children = gtk_container_children(GTK_CONTAINER(notebook));
	for (unsigned int i = 0; i < g_list_length(children); i++) {
		if (i == (unsigned int) page_num) continue; // ignore for now
		GtkWidget *child = gtk_notebook_get_nth_page(notebook,i);
		if (child == NULL) continue;
		p = (Panel *) gtk_object_get_data((GtkObject *) child,"panel");
		if (p == NULL) continue;	// not a panel, I guess
		if (p->getSelfDestruct()) {	// remove the panel
			g_idle_add(idleNukePanel,(gpointer) p);
		}
	}
	return(0);
}

static gboolean
drop_rcv_callback(GtkWidget *w,GdkDragContext *con,gint /*x*/,gint /*y*/,guint /*time*/)
{
	PanelContainer	*destPC;
	GtkWidget *src_widget;
	Panel	*src_panel;

	destPC = (PanelContainer *) gtk_object_get_data(GTK_OBJECT(w),"panelcontainer");
	if (destPC == NULL) {
		fprintf(stderr,"Cannot get dest panelcontainer for drop!\n");
		return(FALSE);
	}
	src_widget = gtk_drag_get_source_widget(con);
	if (src_widget == NULL) {
		fprintf(stderr,"Cannot get src widget for drop!\n");
		return(FALSE);
	}
	src_panel = (Panel *) gtk_object_get_data(GTK_OBJECT(src_widget),"panel");
	/*
	 *	Locate and nuke previous instance meta info
	 */
	gsMovePanel(src_panel,destPC);
	src_panel->raise();
	return(TRUE);
}


static void
evbox_button_pressed(GtkWidget *w,GdkEventButton *event)
{
	Panel	*p;
	PanelContainer	*pc;

	p = (Panel *) gtk_object_get_data(GTK_OBJECT(w),"panel");
        pc = (PanelContainer *) gsSearchPanelsUpward(p,"PanelContainer");
	/*
	 *	Bring it forward
	 */
	gint pageNum = gtk_notebook_page_num((GtkNotebook *) pc->getNotebook(),
					     p->baseWidget());
	gtk_notebook_set_page((GtkNotebook *) pc->getNotebook(),pageNum);
	/*
	 *	See if it is a third mouse button and menu if so.
	 */
	if (event->button == 3) {
		GtkWidget *popup = gtk_menu_new();
		gtk_widget_show(popup);
		p->doMenu(popup);
		gtk_menu_popup(GTK_MENU(popup),NULL,NULL,NULL,NULL,3,event->time);
	}
}

static void
drag_end_callback(GtkWidget *widget,GdkDragContext *context)
{
	gint	x,y;
	gint	root_x,root_y;
	gint	width,height;
	char	buf[128];

	/*
	 *	Find and calculate root window coordinates for the
	 *	current pointer position.
	 */
	gtk_widget_get_pointer(widget, &x, &y);
	gdk_window_get_root_origin(widget->window,&root_x,&root_y);

        GtkWidget *newRoot = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	sprintf(buf,"%s %s",gessieGetAppName(),gessieGetAppVersion());
        gtk_window_set_title(GTK_WINDOW(newRoot),buf);
	gtk_widget_set_uposition(newRoot,root_x + x,root_y + y);
	gtk_widget_show(newRoot);
        GtkWidget       *rootTable = gtk_table_new(1 /*rows*/,1 /*cols*/,FALSE);
	gtk_widget_show(rootTable);
        gtk_container_add(GTK_CONTAINER(newRoot),rootTable);

        PanelContainer *newRootNotebook = new PanelContainer();
        gtk_table_attach(GTK_TABLE(rootTable),newRootNotebook->baseWidget(),
                         0,1,0,1,
                         (enum GtkAttachOptions)(GTK_FILL|GTK_EXPAND),
                         (enum GtkAttachOptions)(GTK_FILL|GTK_EXPAND),0,0);

		/*
		 *	Do the movin'
		 */
	GtkWidget	*src_widget = gtk_drag_get_source_widget(context);
	if (src_widget == NULL) {
		fprintf(stderr,"Cannot get src widget for drop!\n");
		return;
	}
	Panel *src_panel = (Panel *) gtk_object_get_data(GTK_OBJECT(src_widget),
							 "panel");
	/*
	 *	Set the root window about the same size as the original
	 *	dropped window.
	 */
	width = GTK_WIDGET(src_panel->baseWidget())->allocation.width;
	height = GTK_WIDGET(src_panel->baseWidget())->allocation.height;
	gtk_widget_set_usize(newRoot,width+10,height+10);

	gsMovePanel(src_panel,newRootNotebook);
	/*
	 *	Should look at getting rid of from window if it is
	 *	standalone and now empty.
	 */
}

static void
drag_begin_callback(GtkWidget *widget,GdkDragContext *context)
{
        GdkBitmap *mask;
        GtkStyle *style = gtk_widget_get_style(theRootWindow);
	GdkPixmap *drag_pixmap = gdk_pixmap_create_from_xpm_d(
						theRootWindow->window,  &mask,
                                              	&style->bg[GTK_STATE_NORMAL],
                                              	PanelDragIconData);
	gtk_drag_set_icon_pixmap(context,gtk_widget_get_colormap(widget),
                            	 drag_pixmap, NULL, -8, -8);
}

void
PanelContainer::add(Panel *p,const gchar *name,gboolean raiseIt,int pos) 
{
	p->setParent((void *) this);		// I am its parent
	GtkWidget	*w = p->baseWidget();
//fprintf(stderr,"adding Panel <%s> 0x%06x parent=0x%06x\n",p->name(),p,p->getParent());
	GtkWidget *toptable = gtk_table_new(1 /*rows*/,1 /*cols*/,FALSE);
    	GtkWidget *event_box = gtk_event_box_new();
	gtk_widget_show(event_box);
	gtk_drag_source_set(event_box,GDK_BUTTON1_MASK,
			    panel_container_target_table, 
			    n_panel_container_targets,
			    (enum GdkDragAction) (GDK_ACTION_MOVE | GDK_ACTION_COPY));
	gtk_signal_connect(GTK_OBJECT(event_box), "drag_begin",
                      	   GTK_SIGNAL_FUNC (drag_begin_callback),
                      	   (gpointer) w);
	gtk_signal_connect(GTK_OBJECT(event_box), "button_press_event",
                      	   GTK_SIGNAL_FUNC (evbox_button_pressed),
                      	   (gpointer) w);
	gtk_signal_connect(GTK_OBJECT(event_box), "drag_end",
                      	   GTK_SIGNAL_FUNC (drag_end_callback),
                      	   (gpointer) w);


	GtkWidget *table = gtk_table_new(1 /*rows*/,2 /*cols*/,FALSE);
	gtk_container_add(GTK_CONTAINER (event_box), table);

	GtkWidget *icon  = getIcon((gchar **) tabIconData);
	gtk_widget_show(icon);

	gtk_table_attach(GTK_TABLE(table),icon,0,1,0,1,
			 (enum GtkAttachOptions)(GTK_FILL),
			 (enum GtkAttachOptions)(GTK_FILL),0,0);

	GtkWidget *label = gtk_label_new(name);
	gtk_widget_show(label);
	gtk_table_attach(GTK_TABLE(table),label,1,2,0,1,
			 (enum GtkAttachOptions)(GTK_FILL),
			 (enum GtkAttachOptions)(GTK_FILL),0,0);
	gtk_widget_show(table);
	gtk_widget_show(toptable);

		/*
		 *	NOTE:
		 *	set the tab toplevel table widget user data
		 *	to be the label widget so that it can later
		 *	be easily set, without having to figure out
		 *	where it is.
		 */
	gtk_object_set_data((GtkObject *) toptable,"label",(gpointer) label);
	gtk_object_set_data((GtkObject *) event_box,"label",(gpointer) label);
		/*
		 *	NOTE:
		 *	set the event box widget user data
		 *	to be the PanelContainer so that it can ascertain
		 *	the panel when dragged and dropped.
		 */
	gtk_object_set_data((GtkObject *) toptable,	"panel",(gpointer) p);
	gtk_object_set_data((GtkObject *) event_box,	"panel",(gpointer) p);
        gtk_object_set_data((GtkObject *) w,		"panel",(gpointer) p);
	gtk_table_attach(GTK_TABLE(toptable),event_box,0,1,0,1,
			 (enum GtkAttachOptions)(GTK_FILL),
			 (enum GtkAttachOptions)(GTK_FILL),0,0);

	if (pos == -1) {
		gtk_notebook_append_page((GtkNotebook *) theBaseWidget, w,
					 toptable);
	}
	else	{
		gtk_notebook_insert_page((GtkNotebook *) theBaseWidget, w,
					 toptable,pos);
	}
	if (raiseIt)
		p->raise();
}
