/*
        va1000_smbus.c - VA1000's PIIX4 SMBus kernel module

        Copyright (c) 2000 Jerry Katzung	<katzung@valinux.com>

	Core SMBus controller code based on lm_sensors' i2c-piix4.c,
	copyright 1998-99 Frodo Looijaard and Philip Edelbrock.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/desc.h>
#include <linux/va1000_ioctl.h>

#define MODULE_NAME		"va1000_smbus"
#define MODULE_VERSION  	"0.1.0"
#define MODULE_DATE		"20000814"
#define MODULE_NAME_LEN		12
#define MODULE_MAJOR_DEV	62
#define MODULE_MINOR_DEV	162

#ifdef CONFIG_X86_IO_APIC
extern void va1000_smbus_intr_fix(void);  /* exported by the kernel */
#endif


#undef PDEBUG
#ifdef VA1000_DEBUG
#define PDEBUG(fmt, args...) printk(KERN_INFO "va1000_smbus: " fmt, ## args)
#else
#define PDEBUG(fmt, args...)
#endif

#undef PDEBUGG
#define PDEBUGG(fmt, args...) /* nothing but a placeholder */



#define MAX_QUEUE_ELEMENTS	40

/* PIIX4 SMBus address offsets */
#define SMBHSTSTS  (0 + piix4_smba)
#define SMBSLVSTS  (1 + piix4_smba)
#define SMBHSTCNT  (2 + piix4_smba)
#define SMBHSTCMD  (3 + piix4_smba)
#define SMBHSTADD  (4 + piix4_smba)
#define SMBHSTDAT0 (5 + piix4_smba)
#define SMBHSTDAT1 (6 + piix4_smba)
#define SMBBLKDAT  (7 + piix4_smba)
#define SMBSLVCNT  (8 + piix4_smba)
#define SMBSHDWCMD (9 + piix4_smba)
#define SMBSLVEVT  (0xA + piix4_smba)
#define SMBSLVDAT  (0xC + piix4_smba)

/* PCI Address Constants */
#define SMBBA     0x090
#define SMBHSTCFG 0x0D2
#define SMBSLVC   0x0D3
#define SMBSHDW1  0x0D4
#define SMBSHDW2  0x0D5
#define SMBREV    0x0D6

/* PIIX4 SMBus interrupt-related defines */
#define IRQOUT9			9	/* interrupt number for smbus cell */

#define SMBHSTSTS_FAILED	0x10
#define SMBHSTSTS_BUS_ERR	0x08
#define SMBHSTSTS_DEV_ERR	0x04
#define SMBHSTSTS_INTER		0x02

#define HOST_INTERRUPTS		0x1E	/* all possible host intr sources */

#define SMBSLVSTS_ALERT		0x20
#define SMBSLVSTS_SHDW2		0x10
#define SMBSLVSTS_SHDW1		0x08
#define SMBSLVSTS_SLAVE1	0x04

#define SLAVE_INTERRUPTS	0x3C	/* all possible slave intr sources */

/* Other settings */
#define MAX_TIMEOUT		500	/* transaction timeout (in ms) */
#define	ENABLE_INT9		1	/* 1: enable interrupts */

/* Bitmasks for "va1k_initialized" */
#define VA1KI_INITIALIZED	0x01
#define VA1KI_REGISTERED	0x02
#define VA1KI_PROC_INSTALLED	0x04
#define VA1KI_MEMORY_ALLOCATED	0x08
#define VA1KI_INTR_CHECKED	0x20
#define VA1KI_INTR_FIXED	0x40
#define VA1KI_VA1000		0x80

#define max(a,b) (((a)>(b))?(a):(b))


typedef struct q_el {
	struct q_el	*next;
	smbus_pkt	packet;
} q_el;

typedef struct {
	int	length;
	q_el	*head;
	q_el	*tail;
} queue;



static int va1k_open(struct inode *ip, struct file *fp);
static int va1k_release(struct inode *ip, struct file *fp);
static int va1k_ioctl(struct inode *ip, struct file *fp,
                        unsigned int cmd, unsigned long arg);
static void enqueue(queue *q, q_el *qel);
static q_el *dequeue(queue *q);

#ifdef CONFIG_PROC_FS
static int va1k_proc_fn(char *buf,char **start,off_t offset,int len,int unused);
#endif


/************************************************/
/*						*/
/*	Global variables			*/
/*						*/
/************************************************/


static int		va1k_initialized = 0;
static int		va1k_flags = 0;

static int		va1k_major = MODULE_MAJOR_DEV;
static int		va1k_minor = MODULE_MINOR_DEV;

static unsigned short	piix4_smba = 0;

static struct pci_dev	*piix4 = NULL;

static char		*va1k_buffer = NULL;

static spinlock_t	va1k_queue_lock = SPIN_LOCK_UNLOCKED;

static queue		va1k_host_queue   = { 0, NULL, NULL };
static queue		va1k_retire_queue = { 0, NULL, NULL };
static queue		va1k_free_queue   = { 0, NULL, NULL };


static struct file_operations va1k_fops = {
        NULL,           /* llseek       */
        NULL,           /* read         */
        NULL,           /* write        */
        NULL,           /* readdir      */
        NULL,           /* poll         */
        va1k_ioctl,
        NULL,           /* mmap         */
        va1k_open,
        NULL,           /* flush        */
        va1k_release,
        NULL,           /* fsync        */
        NULL,           /* fasync       */
        NULL,           /* chk_media_ch */
        NULL,           /* revalidate   */
        NULL            /* lock         */
};


#ifdef CONFIG_PROC_FS
static struct proc_dir_entry  va1k_proc_entry = {
	0,
	MODULE_NAME_LEN, MODULE_NAME,
	S_IFREG | S_IRUGO,
	1, 0, 0,
	0,
	NULL,
	va1k_proc_fn,
};
#endif


#ifdef MODULE
MODULE_AUTHOR("Jerry Katzung <katzung@valinux.com>");
MODULE_DESCRIPTION("VA1000 SMBus host/slave driver");
#endif


/************************************************/
/*						*/
/*	Driver support				*/
/*						*/
/************************************************/


static int piix4_transaction(void)
{
        int temp;

        PDEBUG("Transaction (pre): CNT=%02x, CMD=%02x, ADD=%02x, DAT0=%02x, "
             "DAT1=%02x\n", inb_p(SMBHSTCNT), inb_p(SMBHSTCMD),
             inb_p(SMBHSTADD), inb_p(SMBHSTDAT0), inb_p(SMBHSTDAT1));

        /* Make sure the SMBus host is ready to start transmitting */
        if ((temp = inb_p(SMBHSTSTS)) != 0x00) {
                PDEBUG("SMBus busy (%02x). Resetting... \n", temp);
                outb_p(temp, SMBHSTSTS);
                if ((temp = inb_p(SMBHSTSTS)) != 0x00) {
                        PDEBUG("Reset failed! (%02x)\n", temp);
                        return -1;
                } else {
                        PDEBUG("Reset succeeded!\n");
                }
        }

        /* start the transaction by setting bit 6 */
        outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT);

        return 0;
}


static int is_write(unsigned short addr)
{
	if (addr & 1)
		return 0;
	else
		return 1;
}


static int piix4_start_transaction(smbus_pkt *p)
{
        int	i, len;

#if 1
	PDEBUG("sz st ty ln ad cm d0  1  2  3  4  5  6  7  8  9\n");
	PDEBUG("%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
		p->size, p->status, p->type, p->len, p->addr, p->cmd,
		p->data[0], p->data[1], p->data[2], p->data[3], p->data[4],
		p->data[5], p->data[6], p->data[7], p->data[8], p->data[9]);
#endif

        switch (p->type) {
        case PIIX4_QUICK_WRITE:
        case PIIX4_QUICK_READ:
                outb_p(p->addr, SMBHSTADD);
                break;
        case PIIX4_BYTE_WRITE:
        case PIIX4_BYTE_READ:
                outb_p(p->addr, SMBHSTADD);
                if (is_write(p->addr))
                        outb_p(p->cmd, SMBHSTCMD);
                break;
        case PIIX4_BYTE_DATA_WRITE:
        case PIIX4_BYTE_DATA_READ:
                outb_p(p->addr, SMBHSTADD);
                outb_p(p->cmd,  SMBHSTCMD);
                if (is_write(p->addr))
                        outb_p(p->data[0], SMBHSTDAT0);
                break;
        case PIIX4_WORD_DATA_WRITE:
        case PIIX4_WORD_DATA_READ:
                outb_p(p->addr, SMBHSTADD);
                outb_p(p->cmd,  SMBHSTCMD);
                if (is_write(p->addr)) {
                        outb_p(p->data[0], SMBHSTDAT0);
                        outb_p(p->data[1], SMBHSTDAT1);
                }
                break;
        case PIIX4_BLOCK_WRITE:
        case PIIX4_BLOCK_READ:
                outb_p(p->addr, SMBHSTADD);
                outb_p(p->cmd,  SMBHSTCMD);
                if (is_write(p->addr)) {
                        len = p->len;
                        if (len < 0)
                                len = 0;
                        if (len > 32)
                                len = 32;
                        outb_p(len, SMBHSTDAT0);
                        i = inb_p(SMBHSTCNT);   /* Reset SMBBLKDAT */
                        for (i = 0; i < len; i++)
                                outb_p(p->data[i], SMBBLKDAT);
                }
                break;
        case PIIX4_PROC_CALL:
        default:
		PDEBUG("Bad transaction type: 0x%02X\n", p->type);
                return -1;
        }

        outb_p((p->type & (7<<2)) + (ENABLE_INT9 & 1), SMBHSTCNT);

        if (piix4_transaction())        /* Error in transaction */
                return -1;

        return 0;
}


static int preflight(smbus_pkt *p)
{
	if (p == NULL)
		return PIIX4_INTERNAL_ERROR;

	switch (p->type) {
	case PIIX4_QUICK_WRITE:
	case PIIX4_BYTE_WRITE:
	case PIIX4_BYTE_DATA_WRITE:
	case PIIX4_WORD_DATA_WRITE:
	case PIIX4_BLOCK_WRITE:
		if (p->addr & 0x01)
			return PIIX4_INTERNAL_ERROR; 
		break;

	case PIIX4_QUICK_READ:
	case PIIX4_BYTE_READ:
	case PIIX4_BYTE_DATA_READ:
	case PIIX4_WORD_DATA_READ:
	case PIIX4_BLOCK_READ:
		if ((p->addr & 0x01) == 0)
			return PIIX4_INTERNAL_ERROR;
		break;
		
	case PIIX4_PROC_CALL:
	default:
		return PIIX4_INTERNAL_ERROR;
	}
	
	if (p->len > 10)
		return PIIX4_INTERNAL_ERROR;
	
	return 0;
}


static int va1k_host_transfer(unsigned long arg)
{
        smbus_pkt  	pkt, *p = &pkt;
        smbus_pkt  	*parg = (smbus_pkt *) arg;
        q_el	   	*qel;
	int	   	error = 0;
	unsigned int	flags;
	
	copy_from_user_ret(p, parg, sizeof(smbus_pkt), -EFAULT);
	
        if ((error = preflight(p)) != 0)	/* valid request? */
		return error;

	/***** Lock the queues *****/
	
	spin_lock_irqsave(&va1k_queue_lock, flags);
	
	error = -EBUSY;		/* assume no free queue elements */

	qel = dequeue(&va1k_free_queue);
	
	if (qel != NULL) {
		memcpy(&qel->packet, p, sizeof(smbus_pkt));
        	qel->packet.status = PIIX4_IN_PROGRESS;
        
        	enqueue(&va1k_host_queue, qel);
        
        	if (va1k_host_queue.length == 1)
        	{
			error = piix4_start_transaction(p);

			if (error)
				PDEBUG("Failed to send packet.\n");
        	}
        }

	spin_unlock_irqrestore(&va1k_queue_lock, flags);
	
	/***** Unlock the queues *****/

        return error;
}


static int va1k_retire(unsigned long arg)
{
        smbus_pkt  	pkt, *p = &pkt;
        smbus_pkt  	*parg = (smbus_pkt *) arg;
        q_el	   	*qel;
        int		error = -EAGAIN;  /* assume nothing to retire */
        unsigned int	flags;

	/***** Lock the queues *****/
	
	spin_lock_irqsave(&va1k_queue_lock, flags);
	
	qel = dequeue(&va1k_retire_queue);
	
	if (qel != NULL) {
		memcpy(p, &qel->packet, sizeof(smbus_pkt));
		memset(&qel->packet, 0, sizeof(smbus_pkt));
		enqueue(&va1k_free_queue, qel);
		error = 0;
	}
	
	spin_unlock_irqrestore(&va1k_queue_lock, flags);
	
	/***** Unlock the queues *****/
	
	if (error)		/* nothing to retire, nothing to copy */
		return error;
	
	copy_to_user_ret(parg, p, sizeof(smbus_pkt), -EFAULT);

        return 0;
}


static void enqueue(queue *q, q_el *qel)  /* assumes spinlock is acquired */
{
	if (q->length == 0)
	{
		q->head = qel;
		q->tail = qel;
	}
	else
	{
		(q->tail)->next = qel;
		q->tail = qel;
	}
	
	qel->next = NULL;
	q->length++;
}


static q_el *dequeue(queue *q)  /* assumes spinlock is acquired */
{
	q_el		*qel = NULL;	/* assume it won't work */

	if (q->length == 1)
	{
		qel = q->head;
		q->head = NULL;
		q->tail = NULL;
		qel->next = NULL;
		q->length = 0;
	}
	else if (q->length > 1)
	{
		qel = q->head;
		q->head = qel->next;
		qel->next = NULL;
		q->length--;
	}
	
	return qel;
}


/************************************************/
/*						*/
/*	Interrupt handlers (top and bottom)	*/
/*						*/
/************************************************/


static int piix4_host_handler(unsigned char host_status)
{
	q_el		*qel;
	smbus_pkt	*p;
	unsigned char	i, n;
	int		run_bottom_half = 0;	/* set to run bottom half */
	
	outb(HOST_INTERRUPTS, SMBHSTSTS);	/* clear intr sources */

	qel = dequeue(&va1k_host_queue);	/* get the "active" transfer */
	
	if (qel == NULL)
	{
		PDEBUG("Host interrupt, but no host queue element.\n");
		return run_bottom_half;
	}

	p = &qel->packet;
	
	if (host_status == SMBHSTSTS_INTER)	/* Successful transaction */
	{
		if (is_write(p->addr) || (p->type == PIIX4_QUICK_READ))
			p->status = 0;		/* Nothing to do for writes */
		else
		{
		    switch (p->type) {
		    case PIIX4_BYTE_READ:  /* Hope data isn't in SMBHSTCMD... */
		    case PIIX4_BYTE_DATA_READ:	/* Receive byte */
			p->data[0] = inb(SMBHSTDAT0);
			p->len = 1;
			break;
		    case PIIX4_WORD_DATA_READ:	/* Read word */
			p->data[0] = inb(SMBHSTDAT0);
			p->data[1] = inb(SMBHSTDAT1);
			p->len = 2;
			break;
		    case PIIX4_BLOCK_READ:	/* Block read */
			p->len = inb(SMBHSTDAT0);	/* return actual */
			n = max(p->len, 10);		/* pin to buffer len */
			for ( i = 0 ; i < n ; i++)
				p->data[i] = inb(SMBBLKDAT);
			break;
		    default:
			PDEBUG("Bad host read transaction type\n");
			p->status = PIIX4_INTERNAL_ERROR;
			break;
		    }
		}
 	}
	else
	{	/* Transaction failed */
		PDEBUG("Host transaction failed.\n");
		p->status = PIIX4_IO_ERROR;
	}

	enqueue(&va1k_retire_queue, qel);

	return run_bottom_half;
}


static int piix4_slave_handler(unsigned char slave_status)
{
	q_el		*qel;
	smbus_pkt	*p;
	unsigned char	c;
	unsigned short	w;
	int		run_bottom_half = 0; /* bottom half run if true */
	
	/* Slave interrupt - data is in the PIIX4 slave data register. */

	if (slave_status & SMBSLVSTS_SHDW1)
	{
		/* Slave shadow port 1 interrupt */
		
		c = inb(SMBSHDWCMD);
		w = inw(SMBSLVDAT);

		PDEBUG("Slave Shadow1 event: cmd = 0x%02X, data = 0x%04X\n",
			c, w);
		
		if ((qel = dequeue(&va1k_free_queue)) != NULL)
		{
			p = &qel->packet;
			p->size		= sizeof(smbus_pkt);
			p->status	= slave_status;
			p->type		= PIIX4_SLAVE;
			p->len		= 2;
			p->addr		= 0x10;
			p->cmd		= c;
			p->data[0]	= w & 0x00FF;
			p->data[1]	= (w >> 8) & 0x00FF;
		
			enqueue(&va1k_retire_queue, qel);
		}
		else
			PDEBUG("No free elements to handle slave interrupt\n");
	}

	outb(SLAVE_INTERRUPTS, SMBSLVSTS);	/* Clear slave status bits */

	return run_bottom_half;
}


static void piix4_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	/* snapshots of interrupt states */
	unsigned char   host_status  = inb(SMBHSTSTS);
	unsigned char   slave_status = inb(SMBSLVSTS);
	int		run_bottom_half = 0;
	unsigned int	flags;

	PDEBUG("piix4_isr: host_status = 0x%02X, slave_status = 0x%02X\n",
		host_status, slave_status);

	/***** Lock the queues *****/
	
	spin_lock_irqsave(&va1k_queue_lock, flags);

	if (host_status & HOST_INTERRUPTS)
		run_bottom_half = piix4_host_handler(host_status);
		
	if (slave_status & SLAVE_INTERRUPTS)
		run_bottom_half |= piix4_slave_handler(slave_status);
	
	if (run_bottom_half)
		;
		
	spin_unlock_irqrestore(&va1k_queue_lock, flags);
	
	/***** Unlock the queues *****/
}


/* To work around a BIOS/hardware problem on early VA 1000 units, we check
 * to see that SMBus interrupts are working by issuing a test host request.
 * If the interrupt handler is invoked, then all is well.  Otherwise, we
 * need to whack the IOAPIC to get around the problem.
 */

static void check_and_apply_interrupt_fix(void)
{
	q_el		*qel;
	smbus_pkt	*p;
	int		error;
	unsigned long	timeout;

	if (va1k_initialized & VA1KI_INTR_CHECKED)
		return;		/* been here, done what we can... */

	va1k_initialized |= VA1KI_INTR_CHECKED;	/* remember we were here */

	qel = dequeue(&va1k_free_queue);
	
	if (qel == NULL) {
		PDEBUG("Failed to get a queue element!\n");
		PDEBUG("Interrupts may not work!\n");
		return;
	}
	
	/* Issue an SMBus quick write to an unoccupied address - it
	   will fail, but that's okay - it will generate an interrupt  */

	p = &qel->packet;
	memset(p, 0, sizeof(smbus_pkt));
	p->size		= sizeof(smbus_pkt);
	p->status	= PIIX4_IN_PROGRESS;
	p->type		= PIIX4_QUICK_WRITE;
	p->len		= 0;
	p->addr		= 0x66;		/* should be harmless */
		
	enqueue(&va1k_host_queue, qel);
	error = piix4_start_transaction(p);
	if (error) {
		PDEBUG("Failed to send test packet!\n");
		PDEBUG("Interrupts may not work!\n");
		return;
	}

	/* wait for transaction to show up on retire queue
	   or the interrupt bit to be set in a register  */

	timeout = jiffies;
	PDEBUG("jiffies,timeout = 0x%lx,0x%lx\n", timeout, timeout + 2 * HZ);
	timeout += 2 * HZ;

	while (jiffies < timeout)
		schedule();

	if ((va1k_host_queue.length==0) && (va1k_retire_queue.length==1)) {
		PDEBUG("SMBus interrupt is working properly.\n");
		qel = dequeue(&va1k_retire_queue);
		if (qel) {
			memset(&qel->packet, 0, sizeof(smbus_pkt));
			enqueue(&va1k_free_queue, qel);
		}
	} else {
		unsigned char host_status  = inb(SMBHSTSTS);
		unsigned char slave_status = inb(SMBSLVSTS);

		outb(HOST_INTERRUPTS, SMBHSTSTS);  /* clear any interrupts */
		outb(SLAVE_INTERRUPTS, SMBSLVSTS);

		PDEBUG("No interrupt detected - checking chip directly...\n");
		PDEBUG("   host_status = 0x%02X, slave_status = 0x%02X\n",
			host_status, slave_status);

		if ((host_status & HOST_INTERRUPTS) == 0) {
			PDEBUG("No interrupts detected!?!\n");
			PDEBUG("Interrupts may not work!\n");
			return;
		}

		if ((qel = dequeue(&va1k_host_queue))) {
			memset(&qel->packet, 0, sizeof(smbus_pkt));
			enqueue(&va1k_free_queue, qel);
		}
		
		/* The interrupt didn't get through - whack the IOAPIC */

		PDEBUG("SMBus interrupt handling needs to be fixed.\n");

#ifdef CONFIG_X86_IO_APIC
		va1000_smbus_intr_fix();  /* kernel does the dirty work */
#endif

		va1k_initialized |= VA1KI_INTR_FIXED;
	}

	PDEBUG("<<< check_and_apply_interrupt_fix\n");
}


/************************************************/
/*						*/
/*	Driver methods (open, ioctl, etc.)	*/
/*						*/
/************************************************/


static int va1k_open(struct inode *ip, struct file *fp)
{
	int	error = 0;
	unsigned int	flags;
	
	if (MINOR(ip->i_rdev) != va1k_minor)
		return -ENODEV;

	error = request_irq(IRQOUT9, piix4_isr, 0, MODULE_NAME, NULL);
	if (error)
		return error;
	
	/***** Lock the queues *****/
	
	spin_lock_irqsave(&va1k_queue_lock, flags);

	pci_write_config_byte(piix4, SMBSHDW1, 0x10);	/* our slave address */
	outb(0x0F, SMBSLVCNT);			/* enable all slave intrs */
	
	spin_unlock_irqrestore(&va1k_queue_lock, flags);
	
	/***** Unlock the queues *****/

	check_and_apply_interrupt_fix();
	
	MOD_INC_USE_COUNT;
        return 0;
}


static int va1k_release(struct inode *ip, struct file *fp)
{
	unsigned int	flags;
	
	if (MINOR(ip->i_rdev) != va1k_minor)
		return -ENODEV;

	/***** Lock the queues *****/
	
	spin_lock_irqsave(&va1k_queue_lock, flags);

	outb(0, SMBHSTCNT);		/* disable host and slave interrupts */
	outb(0, SMBSLVCNT);
	
	spin_unlock_irqrestore(&va1k_queue_lock, flags);
	
	/***** Unlock the queues *****/

	free_irq(IRQOUT9, NULL);
	
	MOD_DEC_USE_COUNT;
        return 0;
}


static int va1k_ioctl(struct inode *ip, struct file *fp,
                        unsigned int cmd, unsigned long arg)
{
	if (MINOR(ip->i_rdev) != va1k_minor)
		return -ENODEV;

        switch (cmd) {
        case VA1000_SMBUS_HOST_TRANSFER:
                return va1k_host_transfer(arg);

        case VA1000_SMBUS_RETIRE:
                return va1k_retire(arg);

        default:
                return -EINVAL;
        }
}


/************************************************/
/*						*/
/*	Initialization and cleanup support	*/
/*						*/
/************************************************/

/* Detect and configure the PIIX4, if possible */

static int piix4_setup(void)
{
        piix4 = pci_find_device(PCI_VENDOR_ID_INTEL,            \
                                PCI_DEVICE_ID_INTEL_82371AB_3,  \
                                NULL);

        if (!piix4) return -ENODEV;

        pci_read_config_word(piix4, SMBBA, &piix4_smba);
        piix4_smba &= 0xFFF0;

        if (check_region(piix4_smba, 16)) {
                PDEBUG("I/O space region 0x%x already in use!\n", piix4_smba);
                return -ENODEV;
        }

        request_region(piix4_smba, 16, MODULE_NAME);

        return 0;
}


static int piix4_cleanup(void)
{
        if (va1k_initialized & VA1KI_INITIALIZED) {
        	/* turn off interrupts ??? FIXME */
                release_region(piix4_smba, 16);
        }
        
        if (va1k_initialized & VA1KI_REGISTERED)
		unregister_chrdev(va1k_major, MODULE_NAME);
		
#ifdef CONFIG_PROC_FS
	if (va1k_initialized & VA1KI_PROC_INSTALLED)
		proc_unregister(&proc_root, va1k_proc_entry.low_ino);
#endif
		
	if (va1k_initialized & VA1KI_MEMORY_ALLOCATED)
		kfree(va1k_buffer);

	return 0;
}

static int machine_supported(void)
{
	if (va1k_flags & 0x01)
		return 1;
	else if (va1k_initialized & VA1KI_VA1000)
		return 1;
	else
		return 0;
}


static int is_va1000(void)
{
	struct pci_dev	*p;

	p = pci_find_device(PCI_VENDOR_ID_INTEL,		\
				PCI_DEVICE_ID_INTEL_82443GX_0,	\
				NULL);
	if ((!p) || (p->bus->number != 0) || (p->devfn != 0x0000)) {
		PDEBUG("No 82443GX\n");
		return 0;
	}

	p = pci_find_device(PCI_VENDOR_ID_INTEL,		\
				PCI_DEVICE_ID_INTEL_82371AB_3,	\
				NULL);
	if ((!p) || (p->bus->number != 0) || (p->devfn != 0x003B)) {
		PDEBUG("No PIIX4\n");
		return 0;
	}

	p = pci_find_device(PCI_VENDOR_ID_DEC,			\
				PCI_DEVICE_ID_DEC_21152,	\
				NULL);
	if ((!p) || (p->bus->number != 0) || (p->devfn != 0x00A0)) {
		PDEBUG("No DEC bridge\n");
		return 0;
	}

	p = pci_find_device(PCI_VENDOR_ID_CT,		\
				PCI_DEVICE_ID_CT_69000,	\
				NULL);
	if ((!p) || (p->bus->number != 1) || (p->devfn != 0x0000)) {
		PDEBUG("No C&T F69000\n");
		return 0;
	}

	return 1;
}


static int piix4_init(void)
{
        int	res, size, i;
        q_el	*buf;

        PDEBUG("version %s (%s)\n", MODULE_VERSION, MODULE_DATE);

        if (va1k_initialized) {
                PDEBUG("Oops, piix4_init called a second time!\n");
                return -EBUSY;
        }

	if (is_va1000())
		va1k_initialized |= VA1KI_VA1000;

	if (!machine_supported()) {
		PDEBUG("Unsupported machine.\n");
		return -ENODEV;	/* Not accurate, but gets the point across */
	}

        if ((res = piix4_setup())) {
                PDEBUG("PIIX4 not detected, module not inserted.\n");
                piix4_cleanup();
                return res;
        }

        va1k_initialized |= VA1KI_INITIALIZED;

        res = register_chrdev(va1k_major, MODULE_NAME, &va1k_fops);
        if (res < 0) {
                PDEBUG("register_chrdev failed: 0x%08X\n", res);
                piix4_cleanup();
                return res;		/* be sure to return nonzero */
        } else if (va1k_major == 0)
                va1k_major = res;	/* dynamically assigned */

        va1k_initialized |= VA1KI_REGISTERED;
        
        size = MAX_QUEUE_ELEMENTS * sizeof(smbus_pkt);
        va1k_buffer = kmalloc(size, GFP_KERNEL);
        if (va1k_buffer == NULL) {
        	PDEBUG("failed to allocate buffers\n");
        	piix4_cleanup();
        	return -ENOMEM;
        }
        
        memset(va1k_buffer, 0, size);
        
        /* build up the free queue - host and retire queues start out empty */
        buf = (q_el *) va1k_buffer;
	for ( i = 0 ; i < MAX_QUEUE_ELEMENTS ; i++)
		enqueue(&va1k_free_queue, &buf[i]);
		
        va1k_initialized |= VA1KI_MEMORY_ALLOCATED;

#ifdef CONFIG_PROC_FS
	proc_register(&proc_root, &va1k_proc_entry);
	
	va1k_initialized |= VA1KI_PROC_INSTALLED;
#endif
	
        PDEBUG("PIIX4 SMBus controller detected and initialized\n");
        
        return 0;
}


#ifdef MODULE

int init_module(void)
{
	return piix4_init();
}

void cleanup_module(void)
{
	piix4_cleanup();
}

#endif

/************************************************/
/*						*/
/*	/proc filesystem support		*/
/*						*/
/************************************************/

#if CONFIG_PROC_FS

static int sprint_q_el(char *buf, char *str, q_el *qel)
{
	int	   len = 0;
	smbus_pkt  *p  = &qel->packet;
	
	len += sprintf(buf+len, "%s %02X %02X %02X %02X %02X %02X %02X %02X ",
			str,
			p->size, p->status, p->type, p->len,
			p->addr, p->cmd, p->data[0], p->data[1]);
			
	len += sprintf(buf+len, "%02X %02X %02X %02X %02X %02X %02X %02X\n",
			p->data[2], p->data[3], p->data[4], p->data[5],
			p->data[6], p->data[7], p->data[8], p->data[9]);
	return len;
}

static int va1k_proc_fn(char *buf,char **start,off_t offset,int len,int unused)
{
	int			limit = PAGE_SIZE - 80;
	unsigned long		flags;
	q_el			*qel;
	
	len  = 0;
	
	/***** Lock the queues *****/
	
	spin_lock_irqsave(&va1k_queue_lock, flags);
	
	len += sprintf(buf+len,
			"\nQueue lengths: Host %d, Retire %d, Free %d\n",
			va1k_host_queue.length, va1k_retire_queue.length,
			va1k_free_queue.length);
		
	/* dump the host, retire, and free queues */
	
	len += sprintf(buf+len, "\nQueues:\n");
	len += sprintf(buf+len, "  sz st ty ln ad cm d0 d1 ");
	len += sprintf(buf+len, "d2 d3 d4 d5 d6 d7 d8 d9\n");

	for ( qel = va1k_host_queue.head; qel != NULL ; qel = qel->next)
		len += sprint_q_el(buf+len, "H", qel);			
	
	for ( qel = va1k_retire_queue.head; qel != NULL ; qel = qel->next)
		len += sprint_q_el(buf+len, "R", qel);

	/* for completeness...
	for ( qel = va1k_free_queue.head; qel != NULL ; qel = qel->next)
		len += sprint_q_el(buf+len, "F", qel);
	*/

	spin_unlock_irqrestore(&va1k_queue_lock, flags);
	
	/***** Unlock the queues *****/
	
	len += sprintf(buf+len, "\ninitialized = 0x%02X, major = %d\n",
				va1k_initialized, va1k_major);

	/* if (len < limit) */
	len += sprintf(buf+len, "\nlen = %d, limit = %d\n", len, limit);
	
	return len;
}

#endif


